
class EmptyValueError(ValueError): pass

class UnexpectedValueError(ValueError): pass

class InvalidConfig(Exception): pass

class HttpException(Exception): CODE = 400

class HttpHeaderTooLarge(HttpException): CODE = 431

class HttpPayloadTooLarge(HttpException): CODE = 413

class HttpNotFound(HttpException): CODE = 404

class HttpMethodNotAllowed(HttpException): CODE = 405

class NetworkError(RuntimeError): pass



import dnslib
import logging
import inspect

from dnspooh.server import Server as ServerBase


logger = logging.getLogger(__name__)


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(name)s.%(levelname)s] %(message)s",
    handlers=[logging.StreamHandler()]
)

def dns_request(domain):
    return dnslib.DNSRecord.question(domain)


def dns_answer(request, ip_addr):
    request.add_answer(dnslib.RR(request.q.qname, dnslib.QTYPE.A,
        rdata=dnslib.A(ip_addr)))
    return request


def dns_response(domain, *ip_addrs):
    request = dns_request(domain)
    for ip_addr in ip_addrs:
        dns_answer(request, ip_addr)
    return request


class LoggerMixin:
    def log_call(self):
        func = inspect.stack()[1].function
        cls = self.__class__.__name__
        logger.info('%s.%s() is called', cls, func)


class Server(ServerBase, LoggerMixin):
    def __init__(self):
        pass

    def create_scheduled_task(self, coro, timer, name=None, context=None):
        self.log_call()

    async def bootstrap(self):
        self.log_call()
        return True

    async def handle(self, request, **kwargs):
        dns_answer(request, kwargs['answer'])
        return request


from .rules import RuleIfTest, RuleThenTest, RuleBeforeTest, RuleAfterTest
from .block import BlockTest

__all__ = (RuleIfTest, RuleThenTest, RuleBeforeTest, RuleAfterTest, BlockTest)



import logging
import yaml
import pathlib

try:
    from yaml import CSafeLoader as YAMLLoader
except ImportError:
    from yaml import SafeLoader as YAMLLoader

from .upstream import *
from .proxy import *
from .exceptions import InvalidConfig
from .helpers import parse_addr, flat_dict, RandomInt


CONFIG_FILE = 'config.yml'

DEFAULT_LISTEN_HOST = '0.0.0.0'

LISTEN_ADDRESS = '%s:%d' % (DEFAULT_LISTEN_HOST, DEFAULT_DNS_PORT)

UPSTREAM_TIMEOUT = 5000

HTTP_TIMEOUT = 10000

HTTP_HOST = '127.0.0.1'

CACHE_MAX_SIZE = 4096

CACHE_TTL = 86400

BUILTIN_UPSTREAMS = [
    {
        'name': 'cloudflare-1',
        'type': 'dns',
        'host': '1.1.1.1',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'cloudflare',
            'global',
            'ipv4',
        ]
    },
    {
        'name': 'cloudflare-2',
        'type': 'dns',
        'host': '1.0.0.1',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'cloudflare',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'cloudflare-ipv6-1',
        'type': 'dns',
        'host': '2606:4700:4700::1111',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'cloudflare',
            'global',
            'ipv6',
        ],
    },
    {
        'name': 'cloudflare-ipv6-2',
        'type': 'dns',
        'host': '2606:4700:4700::1001',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'cloudflare',
            'global',
            'ipv6',
        ],
    },
    {
        'name': 'cloudflare-tls',
        'type': 'tls',
        'host': '1.1.1.1',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'cloudflare',
            'global',
            'ipv4',
        ],
        'priority': 1,
    },
    {
        'name': 'cloudflare-https',
        'type': 'https',
        'url': 'https://1.1.1.1/dns-query',
        'groups': [
            'cloudflare',
            'global',
            'ipv4',
        ],
    },{
        'name': 'google-1',
        'type': 'dns',
        'host': '8.8.8.8',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'google',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'google-2',
        'type': 'dns',
        'host': '8.8.4.4',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'google',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'google-tls',
        'type': 'tls',
        'host': '8.8.8.8',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'google',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'google-https',
        'type': 'https',
        'url': 'https://dns.google/dns-query',
        'groups': [
            'google',
            'global',
        ],
    },
    {
        'name': 'alidns-1',
        'type': 'dns',
        'host': '223.6.6.6',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'alidns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'alidns-2',
        'type': 'dns',
        'host': '223.5.5.5',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'alidns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'alidns-ipv6-1',
        'type': 'dns',
        'host': '2400:3200::1',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'alidns',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'alidns-ipv6-2',
        'type': 'dns',
        'host': '2400:3200:baba::1',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'alidns',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'alidns-tls',
        'type': 'tls',
        'host': 'dns.alidns.com',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'alidns',
            'cn',
        ],
    },
    {
        'name': 'alidns-https',
        'type': 'https',
        'url': 'https://dns.alidns.com/dns-query',
        'groups': [
            'alidns',
            'cn',
        ],
    },
    {
        'name': '114dns-1',
        'type': 'dns',
        'host': '114.114.114.114',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            '114dns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': '114dns-2',
        'type': 'dns',
        'host': '114.114.115.115',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            '114dns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'onedns-1',
        'type': 'dns',
        'host': '117.50.10.10',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'onedns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'onedns-2',
        'type': 'dns',
        'host': '52.80.52.52',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'onedns',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'onedns-ipv6-1',
        'type': 'dns',
        'host': '2400:7fc0:849e:200::4',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'onedns',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'onedns-ipv6-2',
        'type': 'dns',
        'host': '2404:c2c0:85d8:901::4',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'onedns',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'dnspod',
        'type': 'dns',
        'host': '119.29.29.29',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'dnspod',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'dnspod-tls-1',
        'type': 'tls',
        'host': '120.53.53.53',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'dnspod',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'dnspod-tls-2',
        'type': 'tls',
        'host': '1.12.12.12',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'dnspod',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'dnspod-tls-3',
        'type': 'tls',
        'host': 'dot.pub',
        'port': DEFAULT_DOT_PORT,
        'disable': True,
        'groups': [
            'dnspod',
            'cn',
        ],
    },
    {
        'name': 'dnspod-ipv6',
        'type': 'dns',
        'host': '2402:4e00::',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'dnspod',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'dnspod-https-1',
        'type': 'https',
        'url': 'https://1.12.12.12/dns-query',
        'groups': [
            'dnspod',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'dnspod-https-2',
        'type': 'https',
        'url': 'https://120.53.53.53/dns-query',
        'groups': [
            'dnspod',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'dnspod-https-3',
        'type': 'https',
        'url': 'https://doh.pub/dns-query',
        'disable': True,
        'groups': [
            'dnspod',
            'cn',
        ],
    },
    {
        'name': 'baidu',
        'type': 'dns',
        'host': '180.76.76.76',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'baidu',
            'cn',
            'ipv4',
        ],
    },
    {
        'name': 'baidu-ipv6',
        'type': 'dns',
        'host': '2400:da00::6666',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'baidu',
            'cn',
            'ipv6',
        ],
    },
    {
        'name': 'opendns-1',
        'type': 'dns',
        'host': '208.67.222.222',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'opendns',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'opendns-2',
        'type': 'dns',
        'host': '208.67.220.220',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'opendns',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'adguard-1',
        'type': 'dns',
        'host': '94.140.14.14',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'adguard',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'adguard-2',
        'type': 'dns',
        'host': '94.140.15.15',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'adguard',
            'global',
            'ipv4',
        ],
    },
    {
        'name': 'adguard-ipv6-1',
        'type': 'dns',
        'host': '2a10:50c0::ad1:ff',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'adguard',
            'global',
            'ipv6',
        ],
    },
    {
        'name': 'adguard-ipv6-2',
        'type': 'dns',
        'host': '2a10:50c0::ad2:ff',
        'port': DEFAULT_DNS_PORT,
        'groups': [
            'adguard',
            'global',
            'ipv6',
        ],
    },
    {
        'name': 'adguard-tls',
        'type': 'tls',
        'host': 'dns.adguard-dns.com',
        'port': DEFAULT_DOT_PORT,
        'groups': [
            'adguard',
            'global',
        ],
    },
]

DEFAULT_CONFIG = {
    'ipv6': False,
    'secure': False,
    'debug': False,
    'output': None,
    'listen': LISTEN_ADDRESS,
    'timeout': UPSTREAM_TIMEOUT,
    'upstreams': BUILTIN_UPSTREAMS,
    'proxy': None,
    'geoip': None,
    'cache': {
        'max_size': CACHE_MAX_SIZE,
        'ttl': CACHE_TTL,
    },
    'log': {
        'path': 'access.log',
        'trace': True,
        'payload': True,
    },
    'http': {
        'host': HTTP_HOST,
        'port': RandomInt(1024, 65535),
        'timeout': HTTP_TIMEOUT,
        'disable': False,
    },
    'middlewares': ['cache'],
}


logger = logging.getLogger(__name__)


_parse_addr = functools.partial(parse_addr, DEFAULT_LISTEN_HOST, DEFAULT_DNS_PORT)


def _load_from_file(file_path):
    if not file_path.is_file():
        raise FileNotFoundError('Cannot load config file "%s"' % (file_path.absolute(), ))

    def yaml_include(loader, node):
        node_value = loader.construct_scalar(node)
        with file_path.parent.joinpath(node_value).open() as stream:
            return yaml.load(stream, Loader=YAMLLoader)

    def yaml_path(loader, node):
        node_value = loader.construct_scalar(node)
        return str(file_path.parent.joinpath(node_value).absolute())

    YAMLLoader.add_constructor('!include', yaml_include)
    YAMLLoader.add_constructor('!path', yaml_path)
    with file_path.open() as stream:
        return yaml.load(stream, Loader=YAMLLoader)


def _load_from_args(args):
    conf = dict()

    if args.debug is not None:
        conf['debug'] = args.debug

    if args.listen is not None:
        conf['listen'] = args.listen

    if args.timeout is not None:
        conf['timeout'] = args.timeout

    if args.secure is not None:
        conf['secure'] = args.secure

    if args.ipv6 is not None:
        conf['ipv6'] = args.ipv6

    if args.upstreams is not None:
        conf['upstreams'] = args.upstreams

    return conf


def _merge_dict_recursive(original, addition, reserved = []):
    def _unique(l):
        unique_l = list()
        for _ in l:
            if _ not in unique_l:
                unique_l.append(_)
        return unique_l

    def _reserved(k):
        prefix = k + '.'
        return [_[len(prefix):] for _ in reserved if _.startswith(prefix)]

    for k, v in addition.items():
        if k not in original:
            if v is not None:
                original[k] = v
        elif k in reserved:
            continue
        else:
            orig_v = original[k]
            if isinstance(orig_v, dict) and isinstance(v, dict):
                _merge_dict_recursive(orig_v, v, _reserved(k))
            elif isinstance(orig_v, list) and isinstance(v, list):
                original[k] = _unique(orig_v + v)

    return original


class Config:
    def __init__(self, conf):
        self.conf = conf

    def _get(self, key):
        if key in self.conf:
            return self.conf[key]
        key_nodes = key.split('.')
        value = self.conf
        for key_node in key_nodes:
            value = value[key_node]
        return value

    def get(self, key, default=None):
        try:
            return self._get(key)
        except (KeyError, TypeError):
            return default

    def __getitem__(self, key):
        try:
            return self._get(key)
        except (KeyError, TypeError):
            logger.warning('Configure item "%s" not found', key)

    def exists(self, key):
        try:
            self._get(key)
            return True
        except (KeyError, TypeError):
            return False

    @classmethod
    def load(cls, args):
        conf = _load_from_args(args)

        if args.config:
            config_file = pathlib.Path(args.config)
            conf = _merge_dict_recursive(conf, _load_from_file(config_file))
            logger.info('Config file "%s" loaded', config_file.absolute())
        else:
            config_file = pathlib.Path(CONFIG_FILE)
            if config_file.is_file():
               conf = _merge_dict_recursive(conf, _load_from_file(config_file))
               logger.info('Default config file "%s" loaded', config_file.absolute())

        conf = _merge_dict_recursive(conf, DEFAULT_CONFIG, ['upstreams'])

        try:
            conf['proxy'] = parse_proxy(conf.get('proxy'))
            if 'upstreams+' in conf:
                conf['upstreams'].extend(conf['upstreams+'])
                del conf['upstreams+']

            upstreams = map(parse_upstream, conf['upstreams'])
            if 'upstreams_filter' in conf:
                upstreams_filter = conf['upstreams_filter']
                del conf['upstreams_filter']

                def _filter_upstream(upstream):
                    if upstream.disable:
                        return False
                    if 'name' in upstreams_filter and upstream.name in upstreams_filter['name']:
                        return True
                    if 'group' in upstreams_filter:
                        for group_name in upstream.groups:
                            if group_name in upstreams_filter['group']:
                                return True
                    return False
            else:
                _filter_upstream = lambda _: not _.disable
            conf['upstreams'] = list(filter(_filter_upstream, upstreams))

            listen_addrs = conf['listen']
            if isinstance(listen_addrs, str):
                conf['listen'] = [_parse_addr(listen_addrs)]
            else:
                conf['listen'] = [_parse_addr(addr) for addr in listen_addrs]
        except ValueError as exc:
            raise InvalidConfig(exc)

        return cls(conf)

    def to_json(self):
        conf = [{
            'name': k,
            'value': v,
        } for k, v in flat_dict(self.conf)]
        return conf



import logging
import sqlite3
import time

import dnslib

from . import Middleware, Traceback


logger = logging.getLogger(__name__)


CREATE_DATABASE_SQL = \
'''CREATE TABLE IF NOT EXISTS "log" (
    "id"            INTEGER,
    "created_at"    TEXT,
    "elapsed_time"  REAL,
    "qname"         TEXT,
    "qtype"         TEXT,
    "success"       INTEGER,
    "request"       BLOB,
    "response"      BLOB,
    "traceback"     TEXT,
    "error"         TEXT,
    PRIMARY KEY("id" AUTOINCREMENT)
);
'''

INSERT_FAILURE_SQL = \
'''INSERT INTO "log" 
    (created_at, elapsed_time, qname, qtype, success, request, error, traceback) 
    VALUES 
    (:now, :elapsed_time, :qname, :qtype, :success, :request, :error, :traceback)
'''

INSERT_SUCCESS_SQL = \
'''INSERT INTO "log" 
    (created_at, elapsed_time, qname, qtype, success, request, response, traceback) 
    VALUES 
    (:now, :elapsed_time, :qname, :qtype, :success, :request, :response, :traceback)
'''

SELECT_DATASET_SQL = \
'''SELECT id, created_at, elapsed_time, qname, qtype, success, traceback, error 
    FROM "log" ORDER BY id DESC LIMIT :offset, :page_size
'''

QUERY_PAGE_SIZE = 50


class LogMiddleware(Middleware):

    def __init__(self, path, trace, payload):
        self.path = path
        self.trace = trace
        self.payload = payload
        sqlite3.register_adapter(Traceback, self._adapt_traceback)
        with self._open_db() as db:
            db.execute(CREATE_DATABASE_SQL)

    def _adapt_traceback(self, traceback):
        return str(traceback)

    def _open_db(self):
        return sqlite3.connect(self.path)

    def query_dataset(self, page, page_size=QUERY_PAGE_SIZE):
        offset = (page - 1) * page_size
        with self._open_db() as db:
            result = db.execute(SELECT_DATASET_SQL, {
                'offset': offset,
                'page_size': page_size
            })
            field_names = [column[0] for column in result.description]
            return [dict(zip(field_names, row)) for row in result]

    def query_total(self):
        with self._open_db() as db:
            result = db.execute('SELECT COUNT(*) FROM "log"')
            total, = result.fetchone()
            return total

    async def handle(self, request, **kwargs):
        if self.trace:
            kwargs['traceback'] = traceback = Traceback()
        else:
            traceback = None
        params = {
            'now': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            'qname': str(request.q.qname),
            'qtype': dnslib.QTYPE[request.q.qtype],
            'request': request.pack() if self.payload else None,
            'traceback': traceback,
        }

        start_counter = time.perf_counter()
        try:
            response = await super().handle(request, **kwargs)
        except Exception as exc:
            params |= {'success': False, 'error': str(exc), 'elapsed_time': time.perf_counter() - start_counter}
            with self._open_db() as db: db.execute(INSERT_FAILURE_SQL, params)
            raise
        if response is None:
            params |= {'success': False, 'error': 'N/A', 'elapsed_time': time.perf_counter() - start_counter}
            with self._open_db() as db: db.execute(INSERT_FAILURE_SQL, params)
        else:
            params |= {
                'success': True, 
                'response': response.pack() if self.payload else None, 
                'elapsed_time': time.perf_counter() - start_counter
            }
            with self._open_db() as db: db.execute(INSERT_SUCCESS_SQL, params)
        return response



import asyncio
import ssl
import logging

import certifi

from .helpers import Scheme


logger = logging.getLogger(__name__)


class PoolStreamReaderProtocol(asyncio.StreamReaderProtocol):
    def eof_received(self):
        super().eof_received()
        return False

    def data_received(self, data):
        return super().data_received(data)

    def connection_made(self, transport):
        return super().connection_made(transport)

    def connection_lost(self, exc):
        super().connection_lost(exc)
        if self._connection_lost_cb and self.conn:
            self.conn.exc = exc
            self._connection_lost_cb(self.conn)

    def __init__(self, stream_reader, on_connection_lost=None, **kwds):
        self._connection_lost_cb = on_connection_lost
        self.conn = None
        super().__init__(stream_reader, **kwds)


class Connection:
    def __init__(self, name, reader, writer, udp_tunnel=None):
        self.name = name
        self.reader = reader
        self.writer = writer
        self.udp_tunnel = udp_tunnel
        self.exc = None
        self.idle = True
        self._is_wild = True

    def is_wild(self):
        return self._is_wild

    def register(self):
        self._is_wild = False
        self.writer.transport.get_protocol().conn = self

    def __enter__(self):
        self.idle = False
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None or self.is_wild():
            self.writer.transport.close()
        self.idle = True

    def __repr__(self):
        return self.name

    def is_closing(self):
        return self.writer.transport.is_closing()

    def udp_tunnel_enabled(self):
        return self.udp_tunnel is not None

    def close(self):
        return self.writer.transport.close()

    def abort(self):
        return self.writer.transport.abort()
    
    def to_json(self):
        return self.name


class Pool:
    DEFAULT_LIMIT = 2 ** 16
    DEFAULT_SIZE = 2 ** 10

    def __init__(self, loop=None, size=DEFAULT_SIZE):
        self.size = size
        self.total = 0
        self.connections = dict()
        self.loop = asyncio.get_running_loop() if loop is None else loop

    def add(self, conn):
        if self.DEFAULT_SIZE <= self.total:
            return False

        if conn.is_closing():
            raise ConnectionError('Fail to connect "%s"' % (conn.name, ))

        if conn.name in self.connections:
            if conn in self.connections[conn.name]:
                return False
            self.connections[conn.name].add(conn)
        else:
            self.connections[conn.name] = set([conn])
        self.total += 1
        conn.register()
        logger.debug('Add "%s" to connection pool', conn.name)
        return True

    def remove(self, conn):
        if conn.name in self.connections:
            if conn in self.connections[conn.name]:
                self.connections[conn.name].remove(conn)
                self.total -= 1
                return True

        return False

    def get(self, conn_name):
        if conn_name not in self.connections:
            return
        for conn in self.connections[conn_name]:
            if conn.idle:
                return conn
        return

    def on_connection_lost(self, conn):
        self.remove(conn)
        logger.debug('Remove "%s" from connection pool', conn.name)

    async def connect(self, host, port, 
                      scheme=Scheme.TCP, proxy=None, 
                      limit=DEFAULT_LIMIT, pooled=True, **kwds):
        conn_name = _make_conn_name(host, port, scheme, proxy)
        conn = self.get(conn_name)
        if conn:
            return conn

        reader = asyncio.StreamReader(limit=limit, loop=self.loop)
        protocol = PoolStreamReaderProtocol(reader, self.on_connection_lost, loop=self.loop)
        if proxy:
            try:
                transport, _ = await self.loop.create_connection(
                    lambda: protocol, proxy.host, proxy.port, **kwds)
            except OSError as exc:
                raise ConnectionError('Cannot connect to proxy "%s:%d": %s' % (proxy.host, proxy.port, exc))
            writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)
            if scheme == Scheme.UDP:
                conn = Connection(
                    conn_name, reader, writer, 
                    await proxy.make_udp_tunnel(reader, writer, (host, port))
                )
                if pooled: self.add(conn)
                return conn
            if not await proxy.handshake(reader, writer, (host, port)):
                raise ConnectionError('Failed to handshake with proxy "%s"' % (proxy.url, ))
        elif scheme == Scheme.UDP:
            raise ConnectionError('Naked UDP protocol does not supported')
        else:
            try:
                transport, _ = await self.loop.create_connection(
                    lambda: protocol, host, port, **kwds)
            except OSError as exc:
                raise ConnectionError('Cannot connect to server "%s:%d": %s' % (host, port, exc))
            writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)

        if scheme == Scheme.TLS:
            try:
                ssl_context = ssl.create_default_context(cafile=certifi.where())
                transport = await self.loop.start_tls(transport, protocol, ssl_context)
                writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)
            except ssl.SSLError as exc:
                raise ConnectionError('Failed to establish tls connection: %s' % (exc, ))

        conn = Connection(conn_name, reader, writer)
        if pooled: self.add(conn)
        return conn
        
    def dispose(self):
        for conn_set in self.connections.values():
            for conn in conn_set:
                conn.abort()

    def to_json(self):
        return [{
            'name': name,
            'size': len(conn_set)
        } for name, conn_set in self.connections.items()]


def _make_conn_name(host, port, scheme, proxy):
    name = '%s://%s:%d' % (scheme.name.lower(), host, port)
    return name if proxy is None else '%s/%s' % (proxy.url, name)



import asyncio
import base64
import io
import json
import logging
import enum
import email
import email.policy
import mimetypes
import html
import pathlib
import random
import ipaddress
import dnslib

from collections import namedtuple
from http import HTTPStatus
from http.client import HTTPMessage
from urllib.parse import urlencode, urlsplit, parse_qs, quote

from .exceptions import HttpException, HttpHeaderTooLarge, HttpPayloadTooLarge, InvalidConfig
from .helpers import s_addr, Scheme, JsonEncoder


HTTP_VERSION = 'HTTP/1.1'

DEFAULT_HTTP_PORT = 80

DEFAULT_HTTPS_PORT = 443


logger = logging.getLogger(__name__)

ContentType = namedtuple('ContentType', ['media_type', 'charset', 'boundary', 'name'])

ContentDisposition = namedtuple('ContentDisposition', ['type', 'name', 'filename'])

UploadedFile = namedtuple('UploadedFile', ['filename', 'name', 'content_type', 'content'])


class HTTPMethod(str, enum.Enum):
    CONNECT = 'CONNECT'
    DELETE = 'DELETE'
    GET = 'GET'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    POST = 'POST'
    PUT = 'PUT'
    TRACE = 'TRACE'

    def __str__(self):
        return self.value


class HttpMessage(HTTPMessage):
    def get_content_maintype(self):
        maintype = super().get_content_maintype()
        if maintype == 'multipart':
            return ''
        return maintype

    def set_header(self, name, value, **params):
        if self.get(name):
            del self[name]
        return self.add_header(name, str(value), **params)

    def is_closing(self):
        return self.get('Connection', '').lower() == 'close'


class FormData:
    def __init__(self):
        self.data = []
        self.boundary = self._generate_boundary()

    def _generate_boundary(self):
        alphabet = '0123456789qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'
        extra_tail = ''.join(random.sample(alphabet, 16))
        return '----WebKitFormBoundary' + extra_tail

    def get_boundary(self):
        return self.boundary

    def append(self, name, value):
        self.data.append((name, value))
        return self

    def as_bytes(self):
        with io.BytesIO() as fp:
            for item in self.data:
                key, value = item
                fp.write(b'--')
                fp.write(self.boundary.encode())
                fp.write(b'\r\n')
                fp.write(b'Content-Disposition: form-data; name="%s"\r\n' %(html.escape(key).encode(), ))
                fp.write(b'\r\n')
                fp.write(html.escape(value).encode())
                fp.write(b'\r\n')
            fp.write(b'--')
            fp.write(self.boundary.encode())
            fp.write(b'--\r\n')
            fp.seek(0)
            return fp.read(-1)


class Request:
    def get(self, name, default=None):
        if self.query and name in self.query:
            return self.query[name][0]
        if self.form and name in self.form:
            return self.form[name][0]
        return default
    
    def get_int(self, name, default=0):
        try:
            return int(self.get(name))
        except TypeError:
            return default

    def get_list(self, name, default=[]):
        value = self.get(name)
        if isinstance(value, list):
            return value
        elif value is not None:
            return [value]
        return default

    def get_all(self, name):
        if self.query and name in self.query:
            return self.query[name]
        if self.form and name in self.form:
            return self.form[name]
        return []
    
    def is_json(self):
        return self._is_json

    def json(self):
        if not self.is_json():
            return
        try:
            return json.loads(self._body)
        except json.JSONDecodeError:
            pass

    @property
    def body(self):
        return self._body

    def _parse_post_data(self, body):
        raw_content_type = self.headers.get('Conetent-Type')
        with io.BytesIO() as fp:
            fp.write(b'Content-Type: ')
            fp.write(raw_content_type.encode())
            fp.write(b'\r\n\r\n')
            fp.write(body)
            form_data = email.message_from_binary_file(
                fp, _class=HttpMessage, policy=email.policy.HTTP).get_payload()
        if not form_data: return
        form = dict()
        files = dict()
        for item in form_data:
            content_disposition = parse_content_disposition(item.get('Content-Disposition'))
            if content_disposition.type == 'form-data' and content_disposition.name:
                if content_disposition.filename:
                    uploaded_file = parse_uploaded_file(item, content_disposition)
                    if content_disposition.name in files:
                        files[content_disposition.name].append(uploaded_file)
                    else:
                        files[content_disposition.name] = [uploaded_file]
                else:
                    if content_disposition.name in form:
                        form[content_disposition.name].append(item.get_payload())
                    else:
                        form[content_disposition.name] = [item.get_payload()]
        self.form = form
        self.files = files

    @body.setter
    def body(self, value):
        if value is None: return
        if not isinstance(value, bytes):
            raise TypeError('Response body must be bytes type, %s given' % (type(value), ))

        raw_content_type = self.headers.get('Content-Type')
        content_type = parse_content_type(raw_content_type)
        media_type = content_type.media_type
        charset = content_type.charset
        if media_type == 'applicaton/x-www-urlencoded':
            decoded_body = value.decode(charset) if charset \
                else value.decode(charset)
            self.form = parse_qs(decoded_body)
        elif media_type == 'multipart/form-data':
            self._parse_post_data(value)
        elif media_type == 'application/json':
            self._is_json = True
        self._body = value

    def __init__(self):
        self.method = None
        self._url = None
        self.version = HTTP_VERSION
        self.headers = HttpMessage(email.policy.HTTP)
        self._body = None
        self._is_json = False
        self.query_string = None
        self.path = None
        self.query = None
        self.form = None
        self.files = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        parsed_url = urlsplit(value)
        self._url = value
        self.path = quote(parsed_url.path)
        if parsed_url.query:
            self.query_string = parsed_url.query
            self.query = parse_qs(parsed_url.query)
        else:
            self.query_string = ''
            self.query = dict()

    def __repr__(self):
        if self.method is None or self.url is None:
            return '<%s>' % self.__class__.__name__
        return '<%s: %s %r>' % (
            self.__class__.__name__,
            self.method,
            self.url,
        )


class Response:
    def __init__(self, status=HTTPStatus.OK, body=None):
        if not isinstance(status, HTTPStatus):
            status = HTTPStatus(status)
        if isinstance(body, str):
            body = body.encode()
        self._status = status
        self.is_sent = False
        self.reason = self._status.name
        self.version = HTTP_VERSION
        self.headers = HttpMessage(email.policy.HTTP)
        self.body = body

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if not isinstance(value, HTTPStatus):
            value = HTTPStatus(value)
        self._status = value
        self.reason = value.name

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': len(self.body),
            'type': self.headers.get('Content-Type'),
        }


class FileResponse(Response):
    def __init__(self, file, status=HTTPStatus.OK):
        super().__init__(status)
        self.file = file

    @property
    def size(self):
        return self.file.stat().st_size

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': self.size,
            'type': self.headers.get('Content-Type'),
        }


class JsonResponse(Response):
    def __init__(self, payload, status=HTTPStatus.OK):
        super().__init__(status)
        self.body = json.dumps(payload, cls=JsonEncoder).encode()
        self.headers.add_header('Content-Type', 'application/json', charset='utf-8')

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': len(self.body),
            'type': self.headers.get('Content-Type'),
        }


class Client:
    def __init__(self, hostname, conn):
        self.hostname = hostname
        self.conn = conn

    def prepare_headers(self, headers):
        return headers + [
            ('Host', self.hostname),
            ('Accept-Encoding', 'identity'),
            ('Connection', 'keep-alive'),
        ]

    def build_request(self, method, path, query, headers, body=None):
        req = Request()
        req.method = method
        req.url = '%s?%s' % (path, urlencode(query)) if query else path
        for hdr in headers:
            req.headers.add_header(*hdr)
        if isinstance(body, dict):
            req.headers.set_header('Content-Type', 'application/x-www-form-urlencoded')
            body = urlencode(body).encode()
        elif isinstance(body, FormData):
            req.headers.set_header('Content-Type', 'multipart/form-data', boundary=body.get_boundary())
            body = body.as_bytes()
        req.body = body
        return req

    async def _request(self, request):
        writer = self.conn.writer
        start_line = '%s %s %s\r\n' % (request.method, request.url, request.version)
        writer.write(start_line.encode())
        if request.body:
            request.headers.set_header('Content-Length', len(request.body))
            writer.write(request.headers.as_bytes())
            writer.write(request.body)
        else:
            del request.headers['Content-Length']
            writer.write(request.headers.as_bytes())

        try:
            await writer.drain()
        except Exception as exc:
            raise HttpException(str(exc))

    async def _read_response(self):
        resp = Response()
        try:
            start_line, resp.headers, resp.body = await _read_http_message(self.conn.reader)
            resp.version, status, resp.reason = start_line.split(' ', 2)
            resp.status = HTTPStatus(int(status))
        except (HttpException, asyncio.CancelledError) as exc:
            raise exc
        except:
            raise HttpException('Invalid response')
        return resp

    async def get(self, path, query=None, headers=[]):
        request = self.build_request(HTTPMethod.GET, path, query, self.prepare_headers(headers))
        logger.debug('Request sent to "%s": %s', self.hostname, request)
        await self._request(request)
        response = await self._read_response()
        logger.debug('Response received from "%s": %s', self.hostname, response)
        return response

    async def post(self, path, query=None, headers=[], body=None):
        request = self.build_request(HTTPMethod.POST, path, query, self.prepare_headers(headers), body)
        logger.debug('Request sent to "%s": %s', self.hostname, request)
        await self._request(request)
        response = await self._read_response()
        logger.debug('Response received from "%s": %s', self.hostname, response)
        return response


class Server:
    MAX_REQUEST_SIZE = 1024 * 1024 * 16

    DEFAULT_FILES = ['index.html', 'index.htm', 'default.htm', 'default.html']

    def __init__(self, config, handler=None, loop=None):
        self.loop = asyncio.get_running_loop() if loop is None else loop
        self.config = config
        self._request_handler = self._create_request_handler(handler)
        self._server = None
        logger.debug('HTTP serivce initialized')
    
    def _create_request_handler(self, handler):
        if hasattr(handler, 'handle'):
            async def _handler(req):
                return await handler.handle(req)
        elif callable(handler):
            async def _handler(req):
                return await handler(req)
        else:
            async def _handler(req):
                return await self.on_error(404)
        return _handler

    async def _respond(self, writer, response):
        try:
            response.headers.add_header('Server', __package__)
            start_line = '%s %d %s\r\n' % (response.version, response.status.value, response.status.phrase)
            writer.write(start_line.encode())
            if isinstance(response, FileResponse):
                ctype, charset = mimetypes.guess_type(response.file)
                kwargs = {} if charset is None else {'charset': charset}
                response.headers.set_header(
                    'Content-Type', 
                    'application/octet-stream' if ctype is None else ctype,
                    **kwargs
                )
                with response.file.open('rb') as fp:
                    response.headers.set_header('Content-Length', response.size)
                    writer.write(response.headers.as_bytes())
                    await writer.drain()
                    await self.loop.sendfile(writer.transport, fp)
                response.is_sent = True
                return response
            elif response.body:
                response.headers.set_header('Content-Length', len(response.body))
                writer.write(response.headers.as_bytes())
                writer.write(response.body)
            else:
                del response.headers['Content-Length']
                writer.write(response.headers.as_bytes())
            await writer.drain()
            response.is_sent = True
        except asyncio.CancelledError:
            raise
        except:
            raise IOError('Response begin sending and an error occurred')
        finally:
            if response.headers.is_closing():
                writer.transport.abort()
        return response
        
    async def _read_request(self, reader):
        req = Request()
        try:
            start_line, req.headers, req.body = await _read_http_message(reader, self.MAX_REQUEST_SIZE)
            method, req.url, req.version = start_line.split(' ', 2)
            req.method = HTTPMethod(method)
        except (HttpException, asyncio.CancelledError):
            raise
        except:
            raise HttpException('Invalid request')
        return req

    def _attempt_static_file(self, path):
        if not self.root:
            return False

        file_path = self.root.joinpath('.' + path).resolve()
        if not str(file_path).startswith(str(self.root)):
            return False

        if file_path.is_dir():
            for default_file in self.DEFAULT_FILES:
                file_path = file_path.joinpath(default_file)
                if file_path.exists():
                    return file_path
        elif file_path.exists():
            return file_path

        return False

    async def on_request(self, request):
        static_file = self._attempt_static_file(request.path)
        if static_file:
            return FileResponse(static_file)
        return await (self._request_handler)(request)

    async def on_error(self, status, body=None):
        resp = Response(status)
        resp.headers.add_header('Connection', 'close')
        resp.body = body
        return resp

    async def on_connect(self, reader, writer):
        peername = writer.transport.get_extra_info('peername')
        logger.debug('Connection from %s', s_addr(peername))
        try:
            while not writer.transport.is_closing():
                try:
                    request = await asyncio.wait_for(self._read_request(reader), self.timeout_sec)
                    logger.debug('Request received from "%s": %s', s_addr(peername), request)
                    response = await self.on_request(request)
                except HttpException as exc:
                    await self._respond(writer, await self.on_error(exc.CODE))
                    break
                except (TimeoutError, asyncio.TimeoutError, EOFError, IOError, asyncio.CancelledError):
                    raise
                except Exception as exc:
                    await self._respond(writer, await self.on_error(HTTPStatus.INTERNAL_SERVER_ERROR))
                    logger.info('Server error on request: %s', exc)
                    break

                try:
                    await self._respond(writer, response)
                    logger.debug('Response sent to "%s": %s', s_addr(peername), response)
                except (TimeoutError, asyncio.TimeoutError, EOFError, IOError, asyncio.CancelledError):
                    raise
                except Exception as exc:
                    logger.info('Server error on response: %s', exc)
                    writer.transport.abort()
                    break
        except (TimeoutError, asyncio.TimeoutError, EOFError, IOError):
            writer.transport.abort()
        except asyncio.CancelledError:
            logger.debug('Connection from %s has been cancelled', s_addr(peername))

    async def run(self):
        http_config = self.config['http']
        if http_config.get('disable'):
            return
        http_root = http_config.get('root')
        if http_root:
            root_path = pathlib.Path(http_root).resolve()
            if not root_path.is_dir():
                raise InvalidConfig('%s is not a directory' % root_path)
            self.root = root_path
        else:
            self.root = None
        host = http_config['host']
        port = int(http_config['port'])
        if host != '127.0.0.1' and host != 'localhost':
            logger.warn('HTTP server host %s is not safe', host)
        self.timeout_sec = http_config['timeout'] / 1000
        self._server = await asyncio.start_server(self.on_connect, host, port)
        logger.info('HTTP serivce started')
        logger.info('HTTP server is available at http://%s:%d/', host, port)

        try:
            async with self._server:
                await self._server.serve_forever()
        except asyncio.CancelledError:
            logger.debug('HTTP serivce interrupted')
        finally:
            self._server.close()
            logger.info('HTTP serivce stopped')


async def _read_http_message(reader, max_body=None):
    try:
        http_header = await reader.readuntil(b'\r\n\r\n')
    except asyncio.exceptions.LimitOverrunError as exc:
        raise HttpHeaderTooLarge(exc)
    with io.StringIO(http_header.decode()) as fp:
        start_line = fp.readline()
        headers = email.message_from_file(fp, _class=HttpMessage, policy=email.policy.HTTP)
    body = None
    if headers.get('Transfer-Encoding') == 'chunked':
        with io.BytesIO() as fp:
            while True:
                chunk_len_ln = await reader.readuntil(b'\r\n')
                chunk_len = int(chunk_len_ln[:-2].decode(), 16)
                if chunk_len == 0: break
                fp.write(await reader.readexactly(chunk_len))
                if b'\r\n' != await reader.readexactly(2):
                    raise HttpException('Chunked encoding error, missing CRLF')
                if max_body and fp.tell() > max_body:
                    raise HttpPayloadTooLarge('The size of payload is greater than %d', max_body)
            fp.seek(0)
            body = fp.read(-1)
    else:
        content_length = headers['Content-Length']
        if content_length is not None:
            with io.BytesIO() as fp:
                body = await reader.readexactly(int(content_length))
        elif headers.get('Connection') == 'close':
            body = await reader.read()
    return start_line, headers, body


def parse_content_type(content_type):
    if not content_type:
        return ContentType(None, None, None, None)
    directives = content_type.split(';')
    media_type = directives.pop(0).strip().lower()
    subitems = dict([[__.strip().lower() for __ in _.split('=', 1)] for _ in directives])

    return ContentType(media_type, 
                       subitems.get('charset'), 
                       subitems.get('boundary'), 
                       subitems.get('name'))


def parse_content_disposition(content_disposition):
    if not content_disposition:
        return ContentDisposition(None, None, None)
    directives = content_disposition.split(';')
    _type = directives.pop(0).strip().lower()
    subitems = dict([[__.strip().lower() for __ in _.split('=', 1)] for _ in directives])

    return ContentType(_type, 
                       subitems.get('name'),  
                       subitems.get('filename'))


def parse_uploaded_file(message, content_disposition=None):
    if content_disposition is None:
        content_disposition = parse_content_disposition(message.get('Content-Disposition'))
    name = html.unescape(content_disposition.name)
    filename = html.unescape(content_disposition.filename)
    encoding = message.get('Content-Transfer-Encoding')
    content_type = parse_content_type(message.get('Conent-Type'))
    content = message.get_payload()
    if encoding == 'base64':
        content = base64.b64decode(content)
    else:
        content = content.encode()
    return UploadedFile(filename, name, content_type, content)


async def fetch(url, resolver, pool, proxy=None, **kwargs):
    parsed_url = urlsplit(url)
    hostname = parsed_url.hostname
    if parsed_url.scheme == 'http':
        scheme = Scheme.TCP
        port = parsed_url.port if parsed_url.port else DEFAULT_HTTP_PORT
    elif parsed_url.scheme == 'https':
        scheme = Scheme.TLS
        port = parsed_url.port if parsed_url.port else DEFAULT_HTTPS_PORT
    else:
        raise ValueError('Invalid url scheme %s' % (parsed_url.scheme, ))
    try:
        ipaddress.ip_address(hostname)
        host = hostname
    except ValueError:
        dns_request = dnslib.DNSRecord.question(hostname)
        dns_response = await resolver(dns_request)
        if not dns_response or dns_response.header.a == 0:
            raise HttpException('Could not resolve domain %s' % (hostname, ))
        host = str(dns_response.rr[0].rdata)
    if 'method' in kwargs:
        method = HTTPMethod(kwargs['method'])
    else:
        method = HTTPMethod.GET
    if method not in (HTTPMethod.GET, HTTPMethod.POST):
        raise ValueError('Unsupported http method %s' % (method.name, ))
    url_path = quote(parsed_url.path)
    if parsed_url.query:
        url_path = '%s?%s' % (url_path, parsed_url.query)
    headers = kwargs.get('headers', [])

    with await pool.connect(host, port, scheme, proxy, pooled=False) as conn:
        if method == HTTPMethod.GET:
            return await Client(hostname, conn).get(url_path, headers=headers)
        else:
            body = kwargs.get('body')
            return await Client(hostname, conn).post(url_path, headers=headers, body=body)


class JSONError(enum.Enum):
    USER_DEFINED = 0, 'User defined error'
    INVALID_JSON = 1, 'Invalid JSON entity'
    ILLEGAL_PARAM = 2, 'Illegal user parameters'

    def __new__(cls, code, message):
        obj = object.__new__(cls)
        obj.code = code
        obj.message = message
        return obj
    
    def to_json(self):
        return {
            'error': {
                'code': self.code,
                'message': self.message,
            }
        }


def json_handler(func):
    def _handle(self, request):
        if not request.is_json():
            raise HttpException()

        try:
            return func(self, **request.json())
        except TypeError:
            return JsonResponse(JSONError.INVALID_JSON, HTTPStatus.BAD_REQUEST)
    return _handle


def async_json_handler(func):
    async def _handle(self, request):
        if not request.is_json():
            raise HttpException()

        try:
            return await func(self, **request.json())
        except TypeError:
            return JsonResponse(JSONError.INVALID_JSON, HTTPStatus.BAD_REQUEST)
    return _handle


def response_json_error(message, code=JSONError.USER_DEFINED.code):
    return JsonResponse({
        'error': {
            'code': code,
            'message': str(message),
        }
    })



import unittest
import pathlib
import tempfile

import dnslib

from dnspooh.middlewares import BlockMiddleware
from . import Server, dns_request, dns_answer, dns_response


server = Server()


async def create_middleware(**kwargs):
    with tempfile.TemporaryDirectory() as path:
        base_path = pathlib.Path(path)
        filenames = []
        for name, content in kwargs.items():
            block_filename = str(base_path.joinpath(name + '.txt'))
            filenames.append(block_filename)
            with open(block_filename, 'w+') as fp:
                fp.write(content)
        block = BlockMiddleware(*filenames)
        block.initialize(server, 'block')
        await block.bootstrap()
    return block


class BlockTest(unittest.IsolatedAsyncioTestCase):
    async def test_block(self):
        block = await create_middleware(
            block1='www.baidu.com', 
            block2='weixin.qq.com',
            block3='ip:127.0.0.1', 
            block4='ip:127.0.0.2'
        )
        request = dns_request('www.baidu.com')
        response = await block.handle(request, answer='127.0.0.1')
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        request = dns_request('weixin.qq.com')
        response = await block.handle(request, answer='127.0.0.1')
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

        request = dns_request('localhost')
        response = await block.handle(request, answer='127.0.0.1')
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        request = dns_request('localhost')
        response = await block.handle(request, answer='127.0.0.2')
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)



import argparse
import asyncio
import logging
import sys

from . import https
from . import server
from . import __version__, __description__
from .config import *
from .upstream import *


logger = logging.getLogger(__name__)


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog = __package__,
        description = __description__,
        add_help=False
    )

    parser.add_argument('-c', '--config', metavar='file', dest='config',
                        help='config file path (example "%s")' % (CONFIG_FILE, ))
    parser.add_argument('-u', '--upstream', metavar='dns_server', dest='upstreams', nargs='+',
                        help='space-separated upstream DNS servers list')
    parser.add_argument('-t', '--timeout', metavar='ms', dest='timeout', type=int, 
                        help='milliseconds for upstream DNS response timeout (default %d ms)' % (UPSTREAM_TIMEOUT, ))
    parser.add_argument('-l', '--listen', metavar='addr', dest='listen', nargs='+', 
                        help='binding to local address and port for DNS proxy server (default "%s")' % (LISTEN_ADDRESS, ))
    parser.add_argument('-o', '--output', metavar='log', dest='output', 
                        help='write stdout to the specified file')
    parser.add_argument('-S', '--secure-only', dest='secure', action='store_true', help='use DoT/DoH upstream servers only')
    parser.add_argument('-6', '--enable-ipv6', dest='ipv6', action='store_true', help='enable IPv6 upstream servers')
    parser.add_argument('-D', '--debug', action='store_true', help='display debug message')
    parser.add_argument('-d', '--dump', action='store_true', help='dump pretty config data')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('-h', '--help', action='help', help='show this help message and exit')

    return parser.parse_args(sys.argv[1:])


async def startup():
    try:
        debug = True
        args = parse_arguments()
        config = Config.load(args)
        if args.dump:
            from pprint import pprint
            pprint(config.conf)
            return
            
        output_file = config.get('output')
        if output_file:
            logging.root.addHandler(logging.FileHandler(output_file))

        debug = config['debug']
        if debug:
            logging.root.setLevel(logging.DEBUG)

        loop = asyncio.get_running_loop()
        loop.set_debug(debug)
        loop.set_exception_handler(lambda _, context: logger.warning(context['message']))

        main_server = server.Server(config, loop)
        http_server = https.Server(config, main_server.handle_http_request, loop)

        await asyncio.gather(main_server.run(), http_server.run())
    except asyncio.CancelledError:
        logger.info('Exit')
    except InvalidConfig as exc:
        logger.error('Invalid config: %s', exc)
    except Exception as exc:
        if debug:
            raise
        else:
            logger.error('Unexpected error: %s', exc)


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(name)s.%(levelname)s] %(message)s",
        handlers=[logging.StreamHandler()]
    )
    try:
        asyncio.run(startup())
    except KeyboardInterrupt:
        pass



import enum
import urllib.parse
import random
import json


def split_domain(name, reverse=True):
    dot_pos = name.find('.')
    if reverse:
        yield name
        while dot_pos > 0:
            yield name[dot_pos + 1:]
            dot_pos = name.find('.', dot_pos + 1)
    else:
        while dot_pos > 0:
            yield name[:dot_pos]
            dot_pos = name.find('.', dot_pos + 1)
        yield name


def s_addr(addr):
    if isinstance(addr, tuple):
        len_addr = len(addr)
        if len_addr == 2:
            return '[%s]:%d' % addr if ':' in addr[0] else '%s:%d' % addr
        elif len_addr == 4:
            return '[%s]:%d' % addr[:2]
    return str(addr)


def parse_addr(default_host, default_port, addr):
    result = urllib.parse.urlsplit('//' + addr)
    return (
        result.hostname or default_host, 
        result.port or default_port
    )


def flat_dict(d, key_prefix='', key_sep='.'):
    flatten = []
    for k, v in d.items():
        _k = key_prefix + key_sep + k
        if isinstance(v, dict):
            flatten.extend(flat_dict(v, _k, key_sep))
        else:
            flatten.append((_k[1:], v))
    return flatten


class RandomInt:
    def __init__(self, begin, end):
        self.begin = begin
        self.end = end

    def __int__(self):
        return random.randrange(self.begin, self.end)
    
    def to_json(self):
        return 'random integer between %d to %d' % (self.begin, self.end)


class Scheme(enum.Enum):
    TCP = enum.auto()
    TLS = enum.auto()
    UDP = enum.auto()


class JsonEncoder(json.JSONEncoder):
    def default(self, o):
        if hasattr(o, 'to_json'):
            return o.to_json()
        
        if hasattr(o, '__iter__'):
            return list(o)

        return super().default(o)



import asyncio
import base64
import logging
import struct
import functools
import enum
import time
import math

from importlib import resources

import maxminddb
import dnslib

from . import https
from . import middlewares
from . import version
from .config import DnsUpstream, HttpsUpstream, TlsUpstream
from .pool import Pool
from .exceptions import *
from .upstream import UpstreamCollection
from .proxy import parse_proxy
from .helpers import s_addr, Scheme


logger = logging.getLogger(__name__)


TEST_DOMAIN = 'www.google.com'


class ServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, server):
        super().__init__()
        self.server = server
        self.need_restart = False

    def datagram_received(self, data, addr):
        self.server.on_request(self.transport, data, addr)

    def error_received(self, exc):
        if isinstance(exc, OSError):
            logger.debug(exc)
        else:
            logger.info('DNS server error received: %s', exc)
        self.need_restart = True
        self.transport.abort()

    def connection_lost(self, exc):
        super().connection_lost(exc)
        if self.need_restart:
            self.need_restart = False
            self.server.on_error_reset(self.transport)

    def connection_made(self, transport):
        self.transport = transport


class QueryProtocol(asyncio.DatagramProtocol):
    def __init__(self, data, response):
        super().__init__()
        self.data = data
        self.response = response

    def connection_made(self, transport):
        self.transport = transport
        transport.sendto(self.data)

    def datagram_received(self, data, addr):
        self.transport.close()
        self.response.set_result(data)

    def error_received(self, exc):
        logger.error('DNS query error: %s', exc)


class Server:
    class Status(enum.Enum):
        INITIALIZED = enum.auto()
        START_PEDDING = enum.auto()
        RUNNING = enum.auto()
        RESTART_PEDDING = enum.auto()
        STOP_PEDDING = enum.auto()
        STOPPED = enum.auto()

    def __init__(self, config, loop=None):
        self.config = config
        self.pool = Pool()
        self.loop = asyncio.get_running_loop() if loop is None else loop
        self.restart_event = asyncio.Event()
        self.status = self.Status.INITIALIZED
        self.tasks = []
        self.transports = []
        logger.debug('DNS serivce initialized')

    async def bootstrap(self):
        logger.debug('DNS service bootstrapping')

        if self.test_geoip('114.114.114.114', 'cn'):
            logger.info('Test GeoIP2 database passed')
        else:
            logger.warning('Test GeoIP2 database failed')

        self.local_addrs = self.config['listen']
        self.timeout_sec = self.config['timeout'] / 1000
        self.upstreams = UpstreamCollection(self.config['upstreams'], 
                                            self.config['secure'],
                                            self.config['ipv6'])
        self.proxy = self.config['proxy']
        if self.proxy:
            logger.info('Using proxy %s', self.proxy)
        bootstrap_upstreams = []
        hostname_upstreams = []
        named_upstreams = dict()

        for upstream in self.upstreams.all():
            if upstream.host:
                bootstrap_upstreams.append(upstream)
            else:
                hostname_upstreams.append(upstream)

            if upstream.name:
                if upstream.name in named_upstreams:
                    raise InvalidConfig('Duplicated upstream name "%s"' % (upstream.name, ))
                named_upstreams[upstream.name] = upstream

        async def resolve_upstream_hostname(hostname_upstream, bootstrap_upstreams):
            request = dnslib.DNSRecord.question(hostname_upstream.hostname)
            response = await self.handle(request, upstreams=bootstrap_upstreams)
            if not response or response.header.a == 0:
                logger.warning('Failed to resolve upstream domain "%s"', hostname_upstream.hostname)
                hostname_upstream.disable = True
            else:
                hostname_upstream.host = str(response.rr[0].rdata)

        await asyncio.gather(*[
            resolve_upstream_hostname(hostname_upstream, bootstrap_upstreams) \
                for hostname_upstream in hostname_upstreams
        ])

        await self.test_all_upstreams(TEST_DOMAIN)
        return True

    async def test_upstream(self, upstream, hostname):
        request = dnslib.DNSRecord.question(hostname)
        start_counter = time.perf_counter()
        response = await self.handle(request, upstreams=[upstream])
        elapsed_time_sec = time.perf_counter() - start_counter
        if not response or response.header.a == 0:
            upstream.disable = True
            upstream.priority = -1
            logger.warning('Test upstream %s failed', upstream.name)
            return False
        timeout_sec = self._get_timeout(upstream)
        upstream.priority = int(max(0, timeout_sec - elapsed_time_sec) / timeout_sec * 1000)
        upstream.disable = False
        logger.info('Test upstream %s passed, responding speed: %d', upstream.name, upstream.priority)
        return True

    async def test_all_upstreams(self, hostname, include_disabled=False):
        queries = [
            self.test_upstream(upstream, hostname) 
            for upstream in self.upstreams.all() 
            if include_disabled or not upstream.disable
        ]
        await asyncio.gather(*queries)
        self.upstreams.sort()
        primary_upstream = self.upstreams.primary
        if primary_upstream.disable:
            raise NetworkError('No available upstream server')
        logger.info('Primary DNS is %s', primary_upstream.name)

    def create_task(self, coro, name=None, context=None):
        task = self.loop.create_task(coro, name=name, context=context)
        self.tasks.append(task)
        task.add_done_callback(lambda _: self.remove_task(task))
        return task

    def create_scheduled_task(self, coro, timer, name=None, context=None):
        async def _task():
            async for i in timer:
                logger.debug('Schedule task "%s" repeat times %d' % (name, i))
                await coro()

        return self.create_task(_task(), name, context)

    def remove_task(self, task):
        self.tasks.remove(task)

    def _get_proxy(self, upstream, **kwargs):
        if 'proxy' in kwargs:
            return parse_proxy(kwargs['proxy'])
        elif upstream.proxy:
            return upstream.proxy
        return self.proxy

    def _create_middlewares(self):
        wrapped = self
        names = self.config['middlewares']
        for name in names:
            try:
                wrapped = middlewares.create_middleware(
                    name, wrapped, self.config.get(name)
                )
            except TypeError as exc:
                raise InvalidConfig(exc)
            logger.info('%s loaded', wrapped.__class__.__name__)
        return wrapped

    async def _resolve_by_dns(self, query, upstream, proxy):
        response_future = self.loop.create_future()
        upstream_addr = upstream.to_addr()

        if proxy and proxy.udp_tunnel_enabled():
            try:
                conn = await self.pool.connect(
                    upstream.host, 
                    upstream.port, 
                    Scheme.UDP, 
                    proxy
                )
            except ConnectionError as exc:
                logger.warning('Failed to connect to proxy %s: %s', proxy.hostname, exc)
                return
            transport, _ = await self.loop.create_datagram_endpoint(
                lambda: QueryProtocol(
                    conn.udp_tunnel.pack(query, upstream_addr), 
                    response_future
                ),
                remote_addr=conn.udp_tunnel.addr
            )
            try:
                response = conn.udp_tunnel.parse(
                    await response_future, 
                    upstream_addr
                )
            finally:
                transport.close()
            return response

        transport, _ = await self.loop.create_datagram_endpoint(
            lambda: QueryProtocol(query, response_future),
            remote_addr=upstream_addr
        )
        try:
            response = await response_future
        finally:
            transport.close()
        return response

    async def _resolve_by_https(self, query, upstream, proxy):
        try:
            with await self.pool.connect(
                upstream.host, 
                upstream.port, 
                Scheme.TLS, 
                proxy
            ) as conn:
                q = base64.b64encode(query).decode().rstrip('=')
                return (await https.Client(upstream.hostname, conn).get(
                    upstream.path,
                    {'dns': q}, 
                    [("Content-type", "application/dns-message")]
                )).body
        except HttpException as exc:
            logger.warning('HTTP exception from %s: %s', upstream.name, exc)
        except ConnectionError as exc:
            logger.warning('Failed to connect to %s: %s', upstream.name, exc)

    async def _resolve_by_tls(self, query, upstream, proxy):
        try:
            with await self.pool.connect(
                upstream.host, 
                upstream.port, 
                Scheme.TLS, 
                proxy
            ) as conn:
                query_size = struct.pack('!H', len(query))
                conn.writer.write(query_size + query)
                await conn.writer.drain()
                response_head = await conn.reader.readexactly(2)
                response_size = struct.unpack('!H', response_head)[0]
                return await conn.reader.readexactly(response_size)
        except asyncio.exceptions.IncompleteReadError:
            logger.error('Failed to read data from %s', upstream.name)
        except ConnectionError as exc:
            logger.warning('Failed to connect to %s: %s', upstream.name, exc)

    async def fetch(self, url, **kwargs):
        return await https.fetch(url, self.handle, 
                                      self.pool, 
                                      self.proxy,
                                      **kwargs)

    def _get_timeout(self, upstream):
        return self.timeout_sec if upstream.timeout_sec is None else upstream.timeout_sec

    def _get_upstreams(self, kwargs):
        customized_upstreams = []
        if 'upstreams' in kwargs:
            customized_upstreams.extend(kwargs['upstreams'])
        if 'upstream' in kwargs:
            customized_upstreams.append(kwargs['upstream'])
        if 'upstream_name' in kwargs:
            upstream_name = kwargs['upstream_name']
            if upstream_name in self.upstreams:
                customized_upstreams.append(self.upstreams[upstream_name])
            else:
                logger.warning('Upstream name %s not defined', upstream_name)
        if 'upstream_group' in kwargs:
            upstream_group = kwargs['upstream_group']
            if self.upstreams.has_group(upstream_group):
                customized_upstreams.extend(self.upstreams.group(upstream_group))
            else:
                logger.warning('Upstream group %s not defined', upstream_group)
        if customized_upstreams:
            kwargs['customized_upstreams'] = True
            return customized_upstreams
        return self.upstreams.sorted

    async def handle(self, request, **kwargs):
        logger.debug('DNS query:\n%s', request)
        data = request.pack()

        for upstream in self._get_upstreams(kwargs):
            proxy = self._get_proxy(upstream, **kwargs)
            if upstream.disable and 'customized_upstreams' not in kwargs:
                continue
            if isinstance(upstream, DnsUpstream):
                resolver = self._resolve_by_dns
            elif isinstance(upstream, HttpsUpstream):
                resolver = self._resolve_by_https
            elif isinstance(upstream, TlsUpstream):
                resolver = self._resolve_by_tls
            else:
                raise TypeError('Invalid upstream type: %s' % (type(upstream).__name__, ))
            if 'traceback' in kwargs:
                kwargs['traceback'].append(upstream.name)

            try:
                upstream.usage += 1
                response_data = await asyncio.wait_for(
                    asyncio.shield(resolver(data, upstream, proxy)), 
                    self._get_timeout(upstream)
                )
                if response_data is None:
                    raise EmptyValueError('Empty response data received')
                try:
                    response = dnslib.DNSRecord.parse(response_data)
                except dnslib.DNSError:
                    raise UnexpectedValueError('Invalid response data received')
                if request.header.id != response.header.id:
                    raise UnexpectedValueError('Response id does not match')
                upstream.success += 1
                logger.debug('DNS response:\n%s', response)
                return response
            except ValueError:
                logger.warning('Failed to resolve by upstream server %s', upstream.name)
            except (TimeoutError, asyncio.exceptions.TimeoutError):
                logger.info('Upstream server %s response timeout', upstream.name)

    async def _handle(self, request):
        resolver = self.middlewares if self.middlewares else self
        if request.header.q > 1:
            coroutines = []
            for q in request.questions:
                _req = request.truncate()
                _req.add_question(q)
                coroutines.append(resolver.handle(_req))
            response = dnslib.DNSRecord(dnslib.DNSHeader(id=request.header.id,
                                           bitmap=request.header.bitmap,
                                           qr=1, ra=1, aa=1),
                                questions=request.questions)
            for _resp in await asyncio.gather(*coroutines):
                if _resp and _resp.header.a > 0:
                    response.add_answer(_resp.a)
            return response
        return await resolver.handle(request)

    def on_response(self, transport, request, addr, future):
        response = future.result()
        if response is None:
            logger.info('Failed to resolve domain name "%s", upstream servers are unreachable', request.q.qname)
            return

        logger.debug('Send response to %s\n%s', s_addr(addr), response)
        try:
            transport.sendto(response.pack(), addr)
        except Exception:
            logger.debug('Failed to send data to %s', s_addr(addr))

    def on_request(self, transport, data, addr):
        request = dnslib.DNSRecord.parse(data)
        logger.debug('Received request from %s\n%s', s_addr(addr), request)
        task = self.loop.create_task(self._handle(request))
        task.add_done_callback(functools.partial(self.on_response, transport, request, addr))

    def restart(self):
        self.status = self.Status.RESTART_PEDDING
        logger.info('Restarting service')
        
        from .cli import parse_arguments
        from .config import Config
        self.config = Config.load(parse_arguments())

        self.restart_event.set()

    def on_error_reset(self, transport):
        self.transports.remove(transport)
        sockname = transport.get_extra_info('sockname')
        async def reset(sockname):
            transport, _ = await self.loop.create_datagram_endpoint(
                lambda: ServerProtocol(self),
                local_addr=sockname
            )
            self.transports.append(transport) 
        return self.loop.create_task(reset(sockname))

    async def run(self):
        self.status = self.Status.START_PEDDING
        await self._run()
        while self.status == self.Status.RESTART_PEDDING:
            self.restart_event.clear()
            await self._run()

    async def _run(self):
        self.middlewares = self._create_middlewares()
        try:
            if not await self.middlewares.bootstrap():
                logger.error('Failed to bootstrap')
                return

            for local_addr in self.local_addrs:
                try:
                    transport, _ = await self.loop.create_datagram_endpoint(
                        lambda: ServerProtocol(self),
                        local_addr=local_addr
                    )
                    self.transports.append(transport)
                    logger.info('DNS service listening on %s', s_addr(local_addr))
                except OSError as exc:
                    logger.error('Failed to start DNS service: %s', exc)
                    return

            self.status = self.Status.RUNNING
            logger.info('DNS serivce started')

            try:
                await self.restart_event.wait()
            except asyncio.CancelledError:
                logger.debug('DNS serivce interrupted')
            finally:
                if self.status == self.Status.RUNNING:
                    self.status = self.Status.STOPPED
                    logger.info('DNS serivce stopped')
        finally:
            for transport in self.transports:
                transport.close()
            self.pool.dispose()

    def open_geoip(self):
        geoip_db = self.config.get('geoip')
        if geoip_db is not None:
            return maxminddb.open_database(geoip_db)

        with resources.open_binary(__package__, 'geoip') as geoip_db:
            return maxminddb.open_database(geoip_db, maxminddb.MODE_FD)
    
    def test_geoip(self, ip, country):
        result = self.open_geoip().get(ip)
        _country = result['country']['iso_code'] if result else None
        return country.upper() == _country

    async def handle_http_request(self, request):
        match request.method, request.path:
            case https.HTTPMethod.GET, '/':
                return https.Response(body='Dnspooh is working')
            case https.HTTPMethod.GET, '/status':
                return https.JsonResponse({
                    'status': self.status.name,
                })
            case https.HTTPMethod.POST, '/restart':
                return https.JsonResponse({
                    'result': self.restart(),
                })
            case https.HTTPMethod.GET, '/version':
                return https.JsonResponse({
                    'version': version.__version__
                })
            case https.HTTPMethod.GET, '/upstreams':
                return https.JsonResponse({
                    'upstreams': self.upstreams,
                    'primary': self.upstreams.primary,
                })
            case https.HTTPMethod.POST, '/upstreams/primary':
                return self._handle_select_primary_upstream(request)
            case https.HTTPMethod.POST, '/upstreams/test':
                return await self._handle_test_upstream(request)
            case https.HTTPMethod.POST, '/upstreams/test-all':
                return await self._handle_test_all_upstream()
            case https.HTTPMethod.GET, '/pool':
                return https.JsonResponse({
                    'pool': self.pool,
                })
            case https.HTTPMethod.GET, '/config':
                return https.JsonResponse({
                    'config': self.config,
                })
            case https.HTTPMethod.GET, '/logs':
                return self._handle_query_access_log(request)
            case https.HTTPMethod.POST, '/dns-query':
                return await self._handle_dns_query(request)
            case https.HTTPMethod.POST, '/geoip2-query':
                return await self._handle_geoip2_query(request)
        raise HttpNotFound()
    
    def _handle_query_access_log(self, request):
        log_middleware = self.middlewares.get_component('log')
        if not isinstance(log_middleware, middlewares.LogMiddleware):
            return https.response_json_error('The log middleware is not enabled')

        page = request.get_int('page', 1)
        total = log_middleware.query_total()
        page_size = middlewares.log.QUERY_PAGE_SIZE
        return https.JsonResponse({
            'total': total,
            'page': {
                'current': page,
                'size': page_size,
                'count': math.ceil(total / page_size),
            },
            'logs': log_middleware.query_dataset(page),
        })

    @https.json_handler
    def _handle_select_primary_upstream(self, name):
        self.upstreams.select(name)
        return https.JsonResponse({
            'result': True,
        })

    @https.async_json_handler
    async def _handle_test_upstream(self, name):
        if name not in self.upstreams:
            return https.JsonResponse(https.JSONError.ILLEGAL_PARAM, 
                                      https.HTTPStatus.BAD_REQUEST) 
        return https.JsonResponse({
            'result': await self.test_upstream(self.upstreams[name], TEST_DOMAIN),
        })

    @https.async_json_handler
    async def _handle_dns_query(self, domain):
        request = dnslib.DNSRecord.question(domain)
        response = await self.handle(request)
        if response is None:
            return https.response_json_error('Failed to resolve domain name %s' % domain)
        return https.JsonResponse({
            'result': str(response)
        })

    @https.async_json_handler
    async def _handle_geoip2_query(self, ip):
        try:
            result = self.open_geoip().get(ip)
        except ValueError as exc:
            return https.response_json_error(exc)
        return https.JsonResponse({
            'result': result
        })

    async def _handle_test_all_upstream(self):
        await self.test_all_upstreams(TEST_DOMAIN, True)
        return https.JsonResponse({
            'result': True,
        })



import logging
import dnslib
import functools
import ipaddress

from . import Middleware, load_config
from .. import timers
from ..exceptions import InvalidConfig
from ..helpers import split_domain


logger = logging.getLogger(__name__)


def _nxdomain(request):
    response = request.reply()
    response.header.rcode = dnslib.RCODE.NXDOMAIN
    return response


def _parse_config(fp):
    domain_blacklist = set()
    ip_blacklist = set()

    for ln in fp:
        ln = ln.strip()
        if ln == '' or ln.startswith('#'):
            continue
        
        if ln.startswith('ip:'):
            ip_blacklist.add(ipaddress.ip_address(ln[3:].strip()))
        else:
            domain_blacklist.add(ln)
    return domain_blacklist, ip_blacklist


class BlockMiddleware(Middleware):
    async def load_config(self, filename):
        try:
            self.blacklists[filename] = await load_config(filename, self.server, _parse_config)
        except Exception as exc:
            logger.warning('Failed to load blacklist "%s": %s', filename, exc)
            return False
        logger.info('Blacklist "%s" loaded', filename)
        return True

    def is_loaded(self, filename):
        return filename in self.blacklists

    async def bootstrap(self):
        if not await super().bootstrap():
            return False
        for _file in self.filenames:
            if isinstance(_file, list):
                filename, refresh_interval = _file
                if not self.is_loaded(filename):
                    self.server.create_scheduled_task(
                        functools.partial(self.load_config, filename),
                        timers.Timer(refresh_interval), 
                        '[SCHEDULE] fetching blacklist %s' % (filename, )
                    )
            else:
                filename = _file
            if self.is_loaded(filename):
                raise InvalidConfig('Duplicate blacklist %s' % (filename, ))
            await self.load_config(filename)
        return True

    def __init__(self, *filenames):
        self.blacklists = dict()
        self.filenames = filenames

    async def handle(self, request, **kwargs):
        if request.q.qtype not in (dnslib.QTYPE.A, dnslib.QTYPE.AAAA):
            return await super().handle(request, **kwargs)

        domain_parts = list(split_domain(request.q.qname.idna().rstrip('.')))
        for blacklist, _ in self.blacklists.values():
            for part in domain_parts:
                if part in blacklist:
                    return _nxdomain(request)

        response = await super().handle(request, **kwargs)
        if response is None: return

        for rr in filter(lambda rr: rr.rtype in (dnslib.QTYPE.A, dnslib.QTYPE.AAAA), response.rr):
            ip_addr = ipaddress.ip_address(str(rr.rdata))
            for _, blacklist in self.blacklists.values():
                if ip_addr in blacklist:
                    return _nxdomain(request)

        return response



#!/usr/bin/env python
import setuptools
import io


pkg_author = 'tabris17'
pkg_name = 'dnspooh'
pkg_url = 'https://github.com/tabris17/dnspooh'
pkg_desc = 'A Lightweight DNS Proxy/Relay'
pkg_requires = [
    'cachetools==5.2.0',
    'certifi==2022.12.7',
    'dnslib==0.9.23',
    'maxminddb==2.2.0',
    'pyparsing==3.0.9',
    'PyYAML==6.0',
]
pkg_version = {}
python_requires = '>=3.10'

with io.open('README.md', 'r', encoding='utf8') as f:
    long_description = f.read()

with io.open('%s/version.py' % (pkg_name, ), 'r', encoding='utf8') as f:
    exec(f.read(), pkg_version)

setuptools.setup(
    author=pkg_author,
    classifiers=[
        "Topic :: Internet :: Name Service (DNS)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    description=pkg_desc,
    include_package_data=True,
    install_requires=pkg_requires,
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=pkg_name,
    packages=setuptools.find_packages(),
    package_dir={'dnspooh': 'dnspooh'},
    package_data={'dnspooh': ['geoip']},
    python_requires=python_requires,
    url=pkg_url,
    entry_points={
        'console_scripts': [
            'dnspooh = dnspooh.cli:main',
        ],
    },
    version=pkg_version['__version__'],
    zip_safe=False
)



import logging
import ipaddress
import dnslib
import functools

from . import Middleware, load_config
from ..exceptions import InvalidConfig
from .. import timers


logger = logging.getLogger(__name__)


def _parse_config(fp):
    hosts = dict()
    line_no = 1

    for ln in fp:
        ln = ln.lstrip()
        if ln == '' or ln.startswith('#'):
            continue
        _addr, _hostname = ln.split(' ', 1)
        addr = _addr.strip()
        hostname = _hostname.strip()
        try:
            ip_addr = ipaddress.ip_address(addr)
        except ValueError:
            raise InvalidConfig('Invalid ip address %s in line %d' % (addr, line_no))
        if hostname in hosts:
            hosts[hostname].append(ip_addr)
        else:
            hosts[hostname] = [ip_addr]
        line_no += 1
    return hosts


class HostsMiddleware(Middleware):
    DEFAULT_TTL = 60

    async def load_config(self, filename):
        try:
            self.hosts[filename] = await load_config(filename, self.server, _parse_config)
        except Exception as exc:
            logger.warning('Failed to load hosts file "%s": %s', filename, exc)
            return False
        logger.info('Hosts file "%s" loaded', filename)
        return True

    def is_loaded(self, filename):
        return filename in self.hosts

    async def bootstrap(self):
        if not await super().bootstrap():
            return False
        for _file in self.filenames:
            if isinstance(_file, list):
                filename, refresh_interval = _file
                self.server.create_scheduled_task(
                    functools.partial(self.load_config, filename),
                    timers.Timer(refresh_interval), 
                    '[SCHEDULE] fetching hosts file %s' % (filename, )
                )
            else:
                filename = _file
            if not self.is_loaded(filename):
                await self.load_config(filename)
            else:
                raise InvalidConfig('Duplicate hosts file %s' % (filename, ))
        return True

    def __init__(self, *filenames):
        self.hosts = dict()
        self.files = []
        self.urls = []
        self.filenames = filenames

    def query(self, request):
        hostname = request.q.qname.idna().rstrip('.')
        qtype = request.q.qtype
        response = request.reply()
        for hosts in self.hosts.values():
            if hostname not in hosts:
                continue
            ip_addrs = hosts[hostname]
            for r in ip_addrs:
                if isinstance(r, ipaddress.IPv6Address) and qtype == dnslib.QTYPE.AAAA:
                    response.add_answer(dnslib.RR(
                        hostname, qtype, 
                        rdata=dnslib.AAAA(str(r)), 
                        ttl=self.DEFAULT_TTL
                    ))
                elif isinstance(r, ipaddress.IPv4Address) and qtype == dnslib.QTYPE.A:
                    response.add_answer(dnslib.RR(
                        hostname, qtype, 
                        rdata=dnslib.A(str(r)), 
                        ttl=self.DEFAULT_TTL
                    ))
        return response if response.header.a > 0 else None

    async def handle(self, request, **kwargs):
        if request.q.qtype not in (dnslib.QTYPE.A, dnslib.QTYPE.AAAA):
            return await super().handle(request, **kwargs)
        response = self.query(request)
        if not response:
            return await super().handle(request, **kwargs)

        return response



import asyncio
import time
import datetime
import re


_timedelta_regex = re.compile(r'((?P<weeks>\d+?)w)?((?P<days>\d+?)d)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?')


def parse_interval(interval):
    parts = _timedelta_regex.match(interval)
    if not parts:
        raise ValueError('%s is not a valid interval' % (interval, ))
    parts = parts.groupdict()
    time_params = {}
    for name, param in parts.items():
        if param:
            time_params[name] = int(param)
    return datetime.timedelta(**time_params).total_seconds()


class Timer:
    def __init__(self, interval, immediately=False, strict=False, limit_times=None):
        if isinstance(interval, str):
            interval = parse_interval(interval)
        self.interval = float(interval)
        self.immediately = immediately
        self.strict = strict
        self.limit_times = limit_times
        self.repeat_times = 0
        self.last_called = None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.limit_times is not None and self.repeat_times > self.limit_times:
            raise StopAsyncIteration()
        delay = self.interval if not self.strict or self.last_called is None else \
            (delay + self.last_called - time.time())
        self.last_called = time.time()
        if self.repeat_times == 0 and self.immediately:
            return self.repeat_times
        self.repeat_times += 1
        await asyncio.sleep(delay)
        return self.repeat_times



import functools

from urllib.parse import urlsplit
from ipaddress import ip_address, IPv6Address


DEFAULT_DOT_PORT = 853

DEFAULT_DNS_PORT = 53

DEFAULT_HTTPS_PORT = 443


class UpstreamCollection:
    def __init__(self, upstreams, only_secure, enable_ipv6):
        if only_secure:
            upstreams = list(filter(
                lambda _: isinstance(_, (TlsUpstream, HttpsUpstream)), 
                upstreams))
        if not enable_ipv6:
            upstreams = list(filter(
                lambda _: not _.host or not isinstance(ip_address(_.host), IPv6Address),
                upstreams))
        if not upstreams:
            raise ValueError('No upstream server available')
        self._upstreams = upstreams
        self._grouped = dict()
        for upstream in upstreams:
            for group in upstream.groups:
                if group in self._grouped:
                    self._grouped[group].append(upstream)
                else:
                    self._grouped[group] = [upstream]
        self._named = {upstream.name: upstream for upstream in upstreams}
        self._sorted = None
        self._selected = None
        self.sort()

    def _cmp(self, up1, up2):
        if up1.name == self._selected: return 1
        if up2.name == self._selected: return -1
        return up1.priority - up2.priority

    def sort(self):
        self._sorted = self.all()
        self._sorted.sort(reverse=True, key=functools.cmp_to_key(self._cmp))
        for group_name, group_upstreams in self._grouped.items():
            group_upstreams = group_upstreams.copy()
            group_upstreams.sort(reverse=True, key=functools.cmp_to_key(self._cmp))
            self._grouped[group_name] = group_upstreams
        return self

    def all(self):
        return self._upstreams.copy()

    def group(self, name):
        return self._grouped[name]

    def has_group(self, name):
        return name in self._grouped

    @property
    def sorted(self):
        if self._sorted is None:
            self.sort()
        return self._sorted

    def select(self, name):
        self._selected = name
        return self.sort()

    @property
    def primary(self):
        return self._sorted[0]

    def __getitem__ (self, name):
        return self._named[name]

    def __contains__(self, name):
        return name in self._named

    def to_json(self):
        return self.sorted


class Upstream:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        if not isinstance(self.name, str) or self.name == '':
            raise ValueError('Upstream name must be a non-empty string')
        self.proxy = kwargs.get('proxy')
        timeout_ms = kwargs.get('timeout')
        self.timeout_sec = timeout_ms / 1000 if timeout_ms is not None else None
        self.groups = kwargs.get('groups', [])
        if not isinstance(self.groups, list):
            raise ValueError('Upstream groups must be a list')
        self.priority = kwargs.get('priority', 0)
        self.success = 0
        self.usage = 0
        self.disable = False

    def __repr__(self):
        return str(self._get_vars())
    
    def to_json(self):
        return self._get_vars()

    def to_addr(self):
        return (self.host, self.port)

    @property
    def health(self):
        if self.usage == 0:
            return -1
        return int((self.success / self.usage) * 100)

    def _get_vars(self):
        _vars = {
            'name': self.name,
            'priority': self.priority,
            'disable': self.disable,
            'health': self.health,
        }
        if self.timeout_sec is not None: _vars['timeout'] = self.timeout_sec
        if self.proxy is not None: _vars['proxy'] = self.proxy
        if self.groups: _vars['groups'] = self.groups
        return _vars


class DnsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.host = kwargs['host']
        self.port = kwargs.get('port', DEFAULT_DNS_PORT)
    
    def _get_vars(self):
        return super()._get_vars() | {
            'host': self.host,
            'port': self.port,
            'type': 'dns',
        }


class HttpsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.url = kwargs['url']
        parsed_url = urlsplit(self.url)
        self.hostname = parsed_url.hostname
        try:
            ip_address(parsed_url.hostname)
            self.host = parsed_url.hostname
        except ValueError:
            self.host = None
        self.port = parsed_url.port if parsed_url.port else DEFAULT_HTTPS_PORT
        self.path = parsed_url.path
    
    def _get_vars(self):
        return super()._get_vars() | {
            'url': self.url,
            'type': 'doh',
        }


class TlsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hostname = kwargs['host']
        try:
            ip_address(self.hostname)
            self.host = self.hostname
        except ValueError:
            self.host = None
        self.port = kwargs.get('port', DEFAULT_DOT_PORT)
    
    def _get_vars(self):
        return super()._get_vars() | {
            'host': self.hostname,
            'port': self.port,
            'type': 'dot',
        }


def parse_upstream(server):
    if isinstance(server, dict):
        server_type = server.get('type', 'https' if 'url' in server else 'dns')
        if server_type == 'dns':
            upstream_class = DnsUpstream
        elif server_type == 'https':
            upstream_class = HttpsUpstream
        elif server_type == 'tls':
            upstream_class = TlsUpstream
        else:
            raise ValueError('Invalid upstream type')
        try:
            return upstream_class(**server)
        except KeyError as e:
            raise ValueError('Missing config key "%s" in "%s"' % (e.args[0], server))
    elif not isinstance(server, str):
        raise TypeError('Parameter server must be dict or string')

    parsed_url = urlsplit(
        server if server.startswith('https://') \
            else '//' + server
    )

    if parsed_url.scheme == 'https':
        return HttpsUpstream(name=server, url=server)

    if parsed_url.path == '' and \
       parsed_url.query == '' and \
       parsed_url.fragment == '':
        if parsed_url.port == DEFAULT_DOT_PORT:
            return TlsUpstream(name=server, host=parsed_url.hostname, port=DEFAULT_DOT_PORT)
        else:
            return DnsUpstream(name=server, 
                               host=parsed_url.hostname, 
                               port=DEFAULT_DNS_PORT \
                                   if parsed_url.port is None \
                                   else parsed_url.port)

    raise ValueError('Invalid upstream format "%s"' % (server, ))
